@@ -5,6 +5,7 @@
 #include <stdio.h>
 
 void *llvm_omp_target_alloc_shared(size_t, int);
+void llvm_omp_target_free_shared(void *, int);
 
 int main() {
   const int N = 64;
@@ -12,8 +13,8 @@ int main() {
 
   int *shared_ptr = llvm_omp_target_alloc_shared(N * sizeof(int), device);
 
-#pragma omp target teams distribute parallel for device(device) \
-            is_device_ptr(shared_ptr)
+#pragma omp target teams distribute parallel for device(device)                \
+    is_device_ptr(shared_ptr)
   for (int i = 0; i < N; ++i) {
     shared_ptr[i] = 1;
   }
@@ -22,8 +23,8 @@ int main() {
   for (int i = 0; i < N; ++i)
     sum += shared_ptr[i];
 
-  omp_target_free(shared_ptr, device);
+  llvm_omp_target_free_shared(shared_ptr, device);
   // CHECK: PASS
   if (sum == N)
-    printf ("PASS\n");
+    printf("PASS\n");
 }
